/**********************************************************************
 * CD Jukebox by Nathan Strong (gblues@jps.net)
 *
 * This software is covered by the GNU Public License. See COPYING
 * file for details on the license. All code contained in these files
 * is (c) 1999 Nathan Strong.
 **********************************************************************/

#include "main.h"
#include "ansi.h"

bool detect_drives(void)
{
  DRIVE_DATA *drive;
  char drvstr[] = "abcd";
  bool found = FALSE;
  int i, j, drvstrlen = strlen(drvstr);

  printf("Searching for IDE/ATAPI CD-ROM drives...\n");
  for( i = 20; i >= 0; i-- )
  {
    for( j = drvstrlen-1; j >= 0; j--)
    {
      char buf[256];
      int fd, err;
      CD_TOCHDR hdr;

      sprintf(buf, "/dev/hd%c%d", drvstr[j], i);
      if( (fd  = open(buf, O_RDONLY)) < 0 )
        continue;
      if( (err = ioctl(fd, CDROMREADTOCHDR, &hdr)) < 0 )
        continue;
      printf("Found valid drive %s!\n", buf);
      drive           = malloc(sizeof *drive);
      drive->filename = strdup(buf + strlen("/dev/"));
      drive->next     = config.drives;
      config.drives   = drive;
      found++;
      close(fd);
    }
  }
  printf("Searching for SCSI CD-ROM drives...\n");
  for(i = 16; i >= 0; i--)
  {
    char buf[256];
    int fd, err;
    CD_TOCHDR hdr;

    sprintf(buf, "/dev/scd%d", i);
    if( (fd  = open(buf, O_RDONLY)) < 0 )
      continue;
    if( (err = ioctl(fd, CDROMREADTOCHDR, &hdr)) < 0 )
      continue;
    printf("Found valid drive %s!\n", buf);
    drive           = malloc(sizeof *drive);
    drive->filename = strdup(buf + strlen("/dev/"));
    drive->next     = config.drives;
    config.drives   = drive;
    found++;
    close(fd);
  }
  return found;
}

void load_tochdrs(void)
{
  int i;
  DRIVE_DATA *dr;
  CD_TOCENTRY entry;
  load_drives();
  for( dr = config.drives; dr != NULL; dr = dr->next)
  {
    int fd;
    CD_TOCHDR *hdr;

    dr->toc    = NULL;
    dr->header = NULL;
    dr->tracks = NULL;
    hdr = malloc( sizeof *hdr );
    if( (fd = open(dr->filename, O_RDONLY)) < 0 )
    {
      perror("read_tochdrs: open");
      free(hdr);
      continue;
    }
    if( ioctl(fd, CDROMREADTOCHDR, hdr) < 0 )
    {
      perror("read_tochdrs: ioctl");
      free(hdr);
      close(fd);
      continue;
    }
    dr->toc = hdr;
    entry.cdte_track  = CDROM_LEADOUT;
    entry.cdte_format = CDROM_MSF;
    if( ioctl(fd, CDROMREADTOCENTRY, &entry) )
    {
      free(hdr);
      dr->toc = NULL;
      close(fd);
      continue;
    }
    dr->header = new_track();
    dr->header->drive  = dr;
    dr->header->length = entry.cdte_addr.msf.minute * 60 +
                         entry.cdte_addr.msf.second;
    dr->header->start  = dr->header->length * 75 +
                         entry.cdte_addr.msf.frame;
    dr->header->frame  = entry.cdte_addr.msf.frame;
    for( i = dr->toc->cdth_trk1 - 1; i >= 0; i-- )
    {
      CD_TRACK *track;
      entry.cdte_track  = i + 1;
      entry.cdte_format = CDROM_MSF;
      if( ioctl(fd, CDROMREADTOCENTRY, &entry) )
      {
        perror("READTOCENTRY: track");
        exit_cdplayer();
      }
      track = new_track();
      track->drive     = dr;
      track->flags     = IS_SET( entry.cdte_ctrl, CDROM_DATA_TRACK );
      track->length    = (entry.cdte_addr.msf.minute * 60) +
                          entry.cdte_addr.msf.second - 1;
      track->start     = track->length * 75 + entry.cdte_addr.msf.frame;
      track->frame     = entry.cdte_addr.msf.frame;
      track->track     = i + 1;
      track->section   = 0;
      track->next      = dr->tracks;
      track->prev      = NULL;
      if( dr->tracks)
        dr->tracks->prev = track;
      dr->tracks       = track;
    }
    if( dr->tracks != NULL )
    {
      CD_TRACK *list;

      for( list = dr->tracks; list != NULL; list = list->next )
      {
        CD_TRACK *nexttrack;
        int start, end;
        nexttrack = list->next ? list->next : dr->header;
        start = (list->start - list->frame) / 75;
        end   = (nexttrack->start - nexttrack->frame) / 75;
        list->length = end - start;
        if( IS_SET(list->flags, CDROM_DATA_TRACK) )
          list->length = (list->next ? list->next->start : dr->header->start) - list->start;
      }
    }
    close(fd);
  }
}

bool do_next_track(void)
{
  if( ui.cur_track->next_playlist == NULL ) return FALSE;
  if( ui.cur_track->next_playlist->drive != ui.cur_track->drive )
    stop_playing();
  ui.cur_track = ui.cur_track->next_playlist;
  ui.cur_cd    = ui.cur_track->drive;
  play_track(ui.cur_track, 0);
  return TRUE;
}

bool do_fast_forward(void)
{
  int pos;

  pos = 10 + (ui.cur_min * 60) + ui.cur_sec;
  if( pos > ui.cur_track->length )
    pos = ui.cur_track->length;

  play_track(ui.cur_track, pos);
  return TRUE;
}

bool do_play(void)
{
  if( ui.playing == PAUSED )
    do_unpause();
  else
    play_track(ui.cur_track, 0);
  return TRUE;
}

bool do_stop(void)
{
  stop_playing();
  return TRUE;
}

bool do_pause(void)
{
  int fd;
  if( (fd = open(ui.cur_cd->filename, O_RDONLY)) < 0)
    return FALSE;
  if( ioctl(fd, CDROMPAUSE, NULL) < 0)
  {
    close(fd);
    return FALSE;
  }
  ui.playing = PAUSED;
  close(fd);
  return TRUE;
}

bool do_rewind(void)
{
  int pos;
  pos = (ui.cur_min * 60) + ui.cur_sec;
  pos = (pos > 10) ? pos - 10 : 0;
  play_track(ui.cur_track, pos);
  return TRUE;
}

bool do_previous_track(void)
{
  if( ui.cur_track->prev_playlist == NULL) return FALSE;
  if( ui.cur_track->drive != ui.cur_track->prev_playlist->drive )
    stop_playing();
  ui.cur_track = ui.cur_track->prev_playlist;
  ui.cur_cd    = ui.cur_track->drive;
  play_track(ui.cur_track, 0);
  return TRUE;
}

void stop_playing(void)
{
  int fd;

  if( (fd = open(ui.cur_cd->filename, O_RDONLY)) < 0)
    return;
  if( (ioctl(fd, CDROMSTOP, NULL)) < 0)
  {
    close(fd);
    return;
  }
  ui.cur_min = ui.cur_sec = 0;
  ui.playing = FALSE;
  SET_BIT(ui.dirty, DIRTY_CLOCK);
  update_ui();
  close(fd);
}  

void play_track(CD_TRACK *track, int offset)
{
  int fd, start = 0, end = 0;
  CD_TRACK *next;

  static struct cdrom_msf trk_zero;
  struct cdrom_msf ind_track;
  struct timeval start_time;

  next = track->next ? track->next : ui.cur_cd->header;
  if( track == NULL || IS_SET(track->flags, CDROM_DATA_TRACK)) return;
  if( (fd = open(ui.cur_cd->filename, O_RDONLY)) < 0)
    return;
  ind_track = trk_zero;
  start = (track->start - track->frame) / 75;
  start++;

  start += offset;
  ui.cur_min = offset / 60;
  ui.cur_sec = offset % 60;

  end = (next->start - next->frame) / 75;

  ind_track.cdmsf_min0   = start / 60;
  ind_track.cdmsf_sec0   = start % 60;
  ind_track.cdmsf_frame0 = track->frame;
  ind_track.cdmsf_min1   = end / 60;
  ind_track.cdmsf_sec1   = end % 60;
  ind_track.cdmsf_frame1 = next->frame;
  if( ioctl(fd, CDROMPLAYMSF, &ind_track) < 0)
    return;
  close(fd);
  gettimeofday(&start_time, NULL);
  current_time = (int) start_time.tv_sec;
  ui.cur_track = track;
  ui.playing   = PLAYING;
  SET_BIT(ui.dirty, DIRTY_TRACK);
  update_ui();
}

void do_unpause(void)
{
  int fd;
  if( ui.playing == PLAYING ) return;

  if( (fd = open(ui.cur_cd->filename, O_RDONLY)) < 0)
    return;
  
  if( ioctl(fd, CDROMRESUME, NULL) < 0)
  {
    close(fd);
    return;
  }
  close(fd);
  ui.playing = PLAYING;
}
